#include "array.h"

NTSTATUS array_alloc(Array *arr, int size);
NTSTATUS array_realloc(Array *arr);
NTSTATUS array_insert(Array *arr, int index, PageInfo *page);
void quicksort(Array *arr, int left, int right);
int partition(Array *arr, int left, int right);
void swap(Array *arr, int index1, int index2);
PageInfo *binary_search(Array *arr, DWORD addr);


NTSTATUS array_init(Array *arr)
{
	NTSTATUS status;
	
	arr->buffer = 0;
	arr->size = 0;
	arr->capacity = PAGE_SIZE;
	
	status = STATUS_UNSUCCESSFUL;
	
	if (array_alloc(arr, arr->capacity) == STATUS_SUCCESS) {
		status = STATUS_SUCCESS;
	}
	
	return status;
}

NTSTATUS array_alloc(Array *arr, int capacity)
{
	NTSTATUS status = STATUS_UNSUCCESSFUL;
	
	if (capacity) {
		arr->buffer = ExAllocatePool(NonPagedPool, capacity);
		if (arr->buffer){
			status = STATUS_SUCCESS;
		}
	}
	
	return status;
}

NTSTATUS array_realloc(Array *arr)
{
	NTSTATUS status;
	PageInfo *OldBuffer;
	int i;
	
	status = STATUS_UNSUCCESSFUL;
	
	if (MmIsAddressValid(arr->buffer)){
		OldBuffer = arr->buffer;
		if (array_alloc(arr, array_size(arr)+PAGE_SIZE) == STATUS_SUCCESS){
			status = STATUS_SUCCESS;
			for (i = 0; i < array_size(arr); ++i) {
				if (MmIsAddressValid(&arr->buffer[i])){
					RtlCopyMemory(&arr->buffer[i], &OldBuffer[i], sizeof(PageInfo));
				}else{
					status = STATUS_UNSUCCESSFUL;
					break;
				}
			}
		}
		ExFreePoolWithTag(OldBuffer, 0);
	}
	
	return status;
}

NTSTATUS array_free(Array *arr)
{
	if (MmIsAddressValid(arr->buffer)){
		ExFreePoolWithTag(arr->buffer, 0);
		arr->buffer = 0;
		return STATUS_SUCCESS;
	}
	return STATUS_UNSUCCESSFUL;
}

NTSTATUS array_insert(Array *arr, int index, PageInfo *page)
{
	if (!MmIsAddressValid((PVOID) page) || !MmIsAddressValid((PVOID) &arr->buffer[index]))
		return STATUS_UNSUCCESSFUL;
		
	RtlCopyMemory(&arr->buffer[index], page, sizeof(PageInfo));
	
	arr->size++;
	
	return STATUS_SUCCESS;
}

NTSTATUS array_insert_back(Array *arr, PageInfo *page)
{
	NTSTATUS status;
	
	status = STATUS_SUCCESS;
	
	if ((arr->capacity - (array_size(arr) * sizeof(PageInfo))) < sizeof(PageInfo)){
		status = array_realloc(arr);
	}
		
	if (status == STATUS_SUCCESS){
		status = array_insert(arr, array_size(arr), page);
	}
		
	return status;
}

PageInfo *array_get_by_index(Array *arr, int index)
{
	if (index >= 0 && index < array_size(arr)){
		return &arr->buffer[index];
	}
		
	return 0;
}

void array_sort(Array *arr)
{
	if (MmIsAddressValid(arr->buffer)){
		quicksort(arr, 0, array_size(arr) - 1);
	}
}

void quicksort(Array *arr, int left, int right)
{
	int mid;
	
	if (left < right) {
		mid = partition(arr, left, right);
		quicksort(arr, left, mid-1);
		quicksort(arr, mid+1, right);
	}
}

int partition(Array *arr, int left, int right)
{
	PageInfo pivot;
	int j;
	int i = left-1;
	
	RtlCopyMemory(&pivot, &arr->buffer[right], sizeof(PageInfo));
	
	for (j = left; j <= right-1; ++j){
		if (arr->buffer[j].dwBaseAddress <= pivot.dwBaseAddress){
			i++;
			swap(arr, i, j);
		}
	}
	swap(arr, i+1, right);
	
	return i+1;
}

void swap(Array *arr, int index1, int index2)
{
	PageInfo temp;
	
	RtlCopyMemory(&temp, &arr->buffer[index1], sizeof(PageInfo));
	RtlCopyMemory(&arr->buffer[index1], &arr->buffer[index2], sizeof(PageInfo));
	RtlCopyMemory(&arr->buffer[index2], &temp, sizeof(PageInfo));
}

PageInfo *array_search(Array *arr, DWORD addr)
{
	if (MmIsAddressValid(arr->buffer)){
		return binary_search(arr, addr);
	}
	
	return 0;
}

PageInfo *binary_search(Array *arr, DWORD addr)
{
	DWORD left, right, middle;
	PageInfo *page;
	
	left = 0;
	right = array_size(arr)-1;
	
	while (left <= right) {
		middle = (left + right) / 2;
		page = array_get_by_index(arr, middle);
		if (page == 0){
			return 0;
		}else if (page->dwBaseAddress == addr){
			return page;
		}else if (addr < page->dwBaseAddress){
			right = middle - 1;
		}else{
			left = middle + 1;
		}
	}
	
	return 0;
}