#include "main.h"
#include "paging.h"
#include "idt.h"
#include "logger.h"

__declspec(dllimport) KAFFINITY KeSetAffinityThread(IN PETHREAD, IN KAFFINITY);

NTSTATUS DriverUnload(IN PDRIVER_OBJECT DriverObject);
NTSTATUS DriverCmd(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);
NTSTATUS LoggerStart(PVOID InputBuffer, ULONG InputBufferLength, PVOID OutputBuffer, ULONG OutputBufferLength);
NTSTATUS LoggerStop(PVOID InputBuffer, ULONG InputBufferLength, PVOID OutputBuffer, ULONG OutputBufferLength);
NTSTATUS DriverDafault(IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIRP);
void NewInt0EHandler(void);
void HookAllCPUs(void);
void UnHookAllCPUs(void);

const WCHAR deviceLinkBuffer[]  = L"\\DosDevices\\page_fault_logger";
const WCHAR deviceNameBuffer[]  = L"\\Device\\page_fault_logger";


PDEVICE_OBJECT g_PagefaultDevice;
void *g_OldInt0EHandler = NULL;
int g_IsInt0EHooked = 0;
KTIMER g_Timer;


/* 
 * Entry Point of driver
 *	
 * Arguments:
 *   Not important
 * 
 * Return:
 *  Not important
 */
NTSTATUS DriverEntry(
   IN PDRIVER_OBJECT  DriverObject,
   IN PUNICODE_STRING RegistryPath )
{
	NTSTATUS ntStatus;
	UNICODE_STRING deviceNameUnicodeString;
	UNICODE_STRING deviceLinkUnicodeString;
	int i;
	
	DbgPrint("Driver Started...\n");
	
	if(IsPAEEnable()){
		DbgPrint("# PAE must be disable\n");
		DbgPrint("# add following line to boot.ini\n");
		DbgPrint("# /noexecute=alwaysoff /nopae\n");
		return STATUS_UNSUCCESSFUL;
	}

	RtlInitUnicodeString (&deviceNameUnicodeString, deviceNameBuffer );
	RtlInitUnicodeString (&deviceLinkUnicodeString, deviceLinkBuffer );
	
	/* for interacting with usermode program */
	ntStatus = IoCreateDevice ( DriverObject,
                                0, 
                                &deviceNameUnicodeString,
                                FILE_DEVICE_PAGEFAULT,
                                0,
                                TRUE,
                                &g_PagefaultDevice );
	
								
	if(! NT_SUCCESS(ntStatus))
	{
        DbgPrint(("Failed to create device!\n"));
        return ntStatus;
    }
	
	/* ... and also symbolic link */
	ntStatus = IoCreateSymbolicLink (&deviceLinkUnicodeString,
                                        &deviceNameUnicodeString );
										
	
	if(! NT_SUCCESS(ntStatus)) 
	{
		IoDeleteDevice(DriverObject->DeviceObject);
        DbgPrint("Failed to create symbolic link!\n");
        return ntStatus;
    }
	
	/* we don't need handle other Irp so we set to deafult routine */
	for(i = 0; i < IRP_MJ_MAXIMUM_FUNCTION; ++i)
	{
		DriverObject->MajorFunction[i] = DriverDafault;
	}
	
	/* when user mode send an IOCTL to our driver we can handle it
	 * with IRP_MJ_DEVICE_CONTROL Irp
	 * and DriverCmd is handling routine
	 */
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = DriverCmd;
	DriverObject->DriverUnload = DriverUnload;

	LoggerStart(NULL, 0, NULL, 0);
	
	return STATUS_SUCCESS;
}

/* 
 * Driver Default routine  for unused Irp
 * 
 * Arguments
 *  Not important 
 * 
 * Return:
 *  None
 */
NTSTATUS DriverDafault(IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIRP)
{
	pIRP->IoStatus.Status = STATUS_SUCCESS;
	pIRP->IoStatus.Information = 0;
	IoCompleteRequest(pIRP,IO_NO_INCREMENT);
	return (STATUS_SUCCESS);
}

/* 
 * Routine called on unloaded driver
 * 
 * Arguments:
 *  Not Important
 *
 * Return:
 *  None
 */
NTSTATUS DriverUnload(IN PDRIVER_OBJECT DriverObject)
{
    UNICODE_STRING deviceLinkUnicodeString;
	PDEVICE_OBJECT p_NextObj;
	KIRQL curIrql;
	
	DbgPrint("Driver Unload...\n");
	
	p_NextObj = DriverObject->DeviceObject;

	if (p_NextObj != NULL)
	{
		RtlInitUnicodeString(&deviceLinkUnicodeString, deviceLinkBuffer);
		IoDeleteSymbolicLink(&deviceLinkUnicodeString);
		IoDeleteDevice(DriverObject->DeviceObject);
		return STATUS_SUCCESS;
	}
	
	return STATUS_SUCCESS;
}

/* 
 * This routine handle IOCTL send by usermode program
 *
 * Arguments:
 *  Not Important
 *
 * Return:
 *  None
 */
NTSTATUS DriverCmd(
    IN PDEVICE_OBJECT DeviceObject, 
    IN PIRP Irp 
    )
{
	PIO_STACK_LOCATION irpStack;
    PVOID inputBuffer;
    PVOID outputBuffer;
    ULONG inputBufferLength;
    ULONG outputBufferLength;
    ULONG ioControlCode;
    NTSTATUS ntstatus;
	
	/* At the begining, assume everything will be ok */
	ntstatus = Irp->IoStatus.Status = STATUS_SUCCESS;
	
	/* we need irp stack */
	Irp->IoStatus.Information = 0;
	irpStack = IoGetCurrentIrpStackLocation(Irp);
	
	/* Get the basic data pointers provided by a user-mode application */
	inputBuffer = Irp->AssociatedIrp.SystemBuffer;
	inputBufferLength = irpStack->Parameters.DeviceIoControl.InputBufferLength;
    outputBuffer = Irp->AssociatedIrp.SystemBuffer;
    outputBufferLength = irpStack->Parameters.DeviceIoControl.OutputBufferLength;
    ioControlCode = irpStack->Parameters.DeviceIoControl.IoControlCode;
	
	switch (ioControlCode){
		case IOCTL_START_LOGGING:
			LoggerStart(inputBuffer, inputBufferLength, outputBuffer, outputBufferLength);
			break;
		case IOCTL_STOP_LOGGING:
			LoggerStart(inputBuffer, inputBufferLength, outputBuffer, outputBufferLength);
			break;
	}
	
	IoCompleteRequest(Irp, IO_NO_INCREMENT);
	return ntstatus;  
}

/* 
 * start logging
 *
 * Arguments:
 *  input/output buffer and their size come/send to usermode/kernelmode
 *
 * Return:
 *  return NTSTATUS value depend on operation
 */
NTSTATUS LoggerStart(
    PVOID InputBuffer, 
    ULONG InputBufferLength, 
    PVOID OutputBuffer, 
    ULONG OutputBufferLength)
{
	HANDLE hProcess;
	NTSTATUS Status;
	LARGE_INTEGER large_int;
	KIRQL curIrql;

	logger_init(1000);
	curIrql = KeRaiseIrqlToDpcLevel();
	if (!g_IsInt0EHooked){
		g_IsInt0EHooked = 1;
		HookAllCPUs();
	}
	KeLowerIrql(curIrql);
	logger_start();

	return STATUS_SUCCESS;
}

NTSTATUS LoggerStop(
    PVOID InputBuffer, 
    ULONG InputBufferLength, 
    PVOID OutputBuffer, 
    ULONG OutputBufferLength)
{
	KIRQL curIrql;

	logger_stop();
	curIrql = KeRaiseIrqlToDpcLevel();
	if (g_IsInt0EHooked){
		g_IsInt0EHooked = 1;
		UnHookAllCPUs();
	}
	KeLowerIrql(curIrql);
	
	return STATUS_SUCCESS;
}

/* 
 * Hook page fault interrupt of all cpu  
 *
 */
void HookAllCPUs(void)
{
	int i;
	PETHREAD CurrentThread;
	
	CurrentThread = PsGetCurrentThread();
	
	for (i = 0; i < KeNumberProcessors; i++){
		KeSetAffinityThread(CurrentThread, 1 << i);
		HookInt(0x0E, NewInt0EHandler, &g_OldInt0EHandler);
	}
}

/* 
 * UnHook page fault interrupt of all cpu  
 *
 */
void UnHookAllCPUs(void)
{
	int i;
	PETHREAD CurrentThread;
	
	CurrentThread = PsGetCurrentThread();
	
	for (i = 0; i < KeNumberProcessors; i++){
		KeSetAffinityThread(CurrentThread, 1 << i);
		UnHookInt(0x0E, g_OldInt0EHandler);
	}
}


/* 
 * Interrupts 0E (page fault) new handler
 *
 *
 *
 */
void __declspec(naked) NewInt0EHandler(void) 
{
	/* +---------------+
	 * |   EFLAGS      | ESP + 0C
	 * +---------------+
	 * |   Return CS   | ESP + 08
	 * +---------------+
	 * |   Return EIP  | ESP + 04
	 * +---------------+
	 * |   Error Code  | ESP + 00
	 * +---------------+ 
	 *	                 ESP points here when handler is executed.,  
	 *					 If there is no error code, ESP points to return EIP
	*/
	
	/* Format of Error Code
	 *
	 *                   4     3      2    1     0     
	 * +-------------------------------------------+
	 * |   Reserved   | I/D | RSVD | U/S | W/R | P |
	 * +-------------------------------------------+
	 *
	 * P   : 0 - note present page; 1 - access violation or use reserved bit
	 * W/R : 0 - read; 1 - write
	 * U/S : 0 - usermode; 1: supervisor
	 * I/D : 0 - fault by not instruction fetch; 1 - fault by instruction fetch (CR4.PAE or IA32_EFER.NXE must be set to 1, if supported)
	 */
	
	__asm {
				pushad
				
				/* ebp point to pushed structure */
				lea ebp, [esp+0x20]
				
				/* address that make page fault */
				mov ebx, cr2
				
				push [ebp]        /* error code */ 
				push ebx          /* fault address */
				push [ebp+0x04]   /* eip */
				push [ebp+0x08]   /* cs */
				movzx eax, _PCR KPCR.Number /* get current processor number */
				push eax          /*  current processor number */
				call logger_enqueue
				
				popad
				jmp g_OldInt0EHandler
	}
}
